<?php
/**
 * Copyright © 2016 Filoblu S.r.l. All rights reserved.
 */

/**
 * Cron Job that:
 * - reads scheduled task from esb_schedule table;
 * - instantiates object taking class/interface from task; 
 * - converts object into configured output format;
 * - calls endpoint;
 * - in case of error schedules again the task.
 */

namespace FiloBlu\Esb\Cron;

use Magento\Framework\Webapi\ServiceOutputProcessor;

class SendMessage
{
    /**
     * @var \FiloBlu\Esb\Model\ScheduleFactory
     */     
    protected $_scheduleFactory;

    /**
     * @var \FiloBlu\Esb\Helper\Format\FormatFactory
     */        
    protected $_formatFactory;

    /**
     * @var \Psr\Log\LoggerInterface
     */         
	protected $_logger;
    
    /**
     * @var \Magento\Framework\ObjectManagerInterface
     */     
    protected $_objectManager;

    /**
     * @var \Magento\Framework\HTTP\ZendClientFactory
     */     
    protected $_httpClientFactory;

    /**
     * @var \Magento\Framework\Webapi\ServiceOutputProcessor
     */     
    protected $_serviceOutputProcessor;

    /**
     * @param \FiloBlu\Esb\Model\ScheduleFactory $scheduleFactory
     * @param \FiloBlu\Esb\Helper\Format\FormatFactory $formatFactory
     * @param \Magento\Framework\HTTP\ZendClientFactory $httpClientFactory
     * @param \Psr\Log\LoggerInterface $logger
     * @param \Magento\Framework\ObjectManagerInterface $objectManager
     * @param ServiceOutputProcessor $serviceOutputProcessor
     */
	public function __construct(
		\FiloBlu\Esb\Model\ScheduleFactory $scheduleFactory,
        \FiloBlu\Esb\Helper\Format\FormatFactory $formatFactory,
        \Magento\Framework\HTTP\ZendClientFactory $httpClientFactory,
        \Psr\Log\LoggerInterface $logger,
        \Magento\Framework\ObjectManagerInterface $objectManager,
        ServiceOutputProcessor $serviceOutputProcessor
	){
		$this->_scheduleFactory=$scheduleFactory;
        $this->_formatFactory=$formatFactory;
        $this->_logger = $logger;
        $this->_objectManager = $objectManager;
        $this->_httpClientFactory = $httpClientFactory;
        $this->_serviceOutputProcessor = $serviceOutputProcessor;
	}

    public function execute()
    {

        /** Retrieve the pending task to be executed */
        $scheduleModel = $this->_scheduleFactory->create();
        $scheduleCollection = $scheduleModel->getCollection()
                                ->addFieldToFilter("status", "pending")
                                ->addFieldToFilter("executable_after",["lteq" => new \Zend_Db_Expr("NOW()")]);

        /** Instantiate format helper (JSON/XML) */
        $formatData = $this->_formatFactory->create();
    
        foreach($scheduleCollection as $schedule){        
            /** Set schedule status to running to prevent concurrent executions.*/
            $schedule->setStatus("running");
            $schedule->setExecutedAt(new \Zend_Db_Expr("NOW()"));
            $schedule->save();
                
            $outputData="";

            /** The scheduled task contains a Repository Interface. Use Repository design pattern to load object. */
            if(!is_null($schedule->getObjectRepositoryInterface())){
                $objectInterface=$schedule->getObjectRepositoryInterface();
                $objectRepository=$this->_objectManager->create($objectInterface);
                $outputObject = $objectRepository->get($schedule->getObjectId());

                /* Use serviceOutputProcessor to build output data. See function processApiRequest() of Magento\Webapi\Controller\Rest. */
                $outputData = $this->_serviceOutputProcessor->process(
                    $outputObject,
                    $objectInterface,
                    "get"
                );                
            }
            /** The scheduled task contains an Object Class. Use standard magento loading. */
            elseif(!is_null($schedule->getObjectClass())){
                $outputObject = $this->_objectManager->create($schedule->getObjectClass())->load($schedule->getObjectId());
                $outputData = $outputObject->getData();
            }
            /** Cannot find interfaces nor classes: cannot instantiate object. */            
            else{
                $schedule->setStatus("permanent_error");
                $schedule->setResponse("Cannot instantiat Object");
                $schedule->save();
                continue;
            }
            $outputData = $formatData->format($outputData);
            //$this->_logger->info("PAYLOAD: ".$outputData);
            $error=false; 

            /** Create HTTP Client. Warning: Zend_HTTP_Client is ZF1 not ZF2.*/
            try{
                $client = $this->_httpClientFactory->create();            
                $client->setUri($schedule->getEndpoint());            
                $client->setRawData($outputData);            
                $client->setEncType($formatData->getContentType());                      
                $client->setMethod(\Zend_Http_Client::POST);
                //$this->_logger->info("Client: ".print_r($client, true));
                $response = $client->request(); 
                //$this->_logger->info("Response: ".print_r($response, true));
                $schedule->setResponse($response->getBody());
                $schedule->setFinishedAt(new \Zend_Db_Expr("NOW()"));
                if($response->isSuccessful()){
                    $schedule->setStatus("success");
                }else{
                    $this->_logger->error($response->getBody());
                    $schedule->setStatus("error");
                    $error=true;
                }
                $schedule->save();
            }catch(Exception $e){
                $this->_logger->error($e->getMessage());
                $schedule->setStatus("error");
                $schedule->setResponse($e->getMessage());
                $schedule->save();
                $error=true;
            }
            if($error){
                $this->scheduleAgain($schedule);
            }
        }
        return $this;
    }

    /**
     * Schedule job again. The schedule time depends on the number of previous errors on the same item.
     *
     * @param  \FiloBlu\Esb\Model\Schedule $schedule
     * @return \FiloBlu\Esb\Model\Schedule $newSchedule
     */
    protected function scheduleAgain($schedule){
        $scheduleModel = $this->_scheduleFactory->create();
        $scheduleCollection = $scheduleModel->getCollection()
                        ->addFieldToFilter("status", "error")
                        ->addFieldToFilter("object_class", $schedule->getObjectClass())
                        ->addFieldToFilter("object_repository_interface", $schedule->getObjectRepositoryInterface())
                        ->addFieldToFilter("object_id", $schedule->getObjectId());
        $errors=$scheduleCollection->getSize();
        if($errors > 0 && $errors < 5){
            $newSchedule = $this->_scheduleFactory->create();
            $minutes = 5 * pow(2 , $errors-1);
            $newSchedule->setData(			
                array(
                    "object_class" => $schedule->getObjectClass(),
                    "object_id" => $schedule->getObjectId(),
                    "endpoint" => $schedule->getEndpoint(),
                    "event" => $schedule->getEvent(),
                    "status" => "pending",
                    "executable_after" => new \Zend_Db_Expr('TIMESTAMPADD(MINUTE,'.$minutes.', NOW())')
			    )
            );
            $newSchedule->save();
            return $newSchedule;
        }else{
            $schedule->setStatus("permanent_error");
            $schedule->save();
            $this->_logger->error("Won't schedule again. More than 4 errors while scheduling. ".json_encode($schedule));
            return false;
        }
    }
}